#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <stdint.h>

/* some enums */
enum formats 	{UNDEF,R_FORM, I_FORM, J_FORM};
enum r_funcs 	{ADD=32,SUB=34};
enum i_ops	{BEQ=4,ADDIU=9,LW=35};
enum j_ops	{J=2};

/* typedefs */
typedef uint32_t Instruction;
typedef uint32_t Address;
typedef enum formats Formats;
typedef enum r_funcs R_Functions;
typedef enum i_ops I_OPCodes;
typedef enum j_ops J_OPCodes;

/* defines for some const's */

//shifts
#define OP_SHIFT 26
#define RS_SHIFT 21
#define RT_SHIFT 16
#define RD_SHIFT 11
#define SH_SHIFT 6
#define FC_SHIFT 0

//masks
#define OP_MASK (uint32_t)(0x3F<<OP_SHIFT)
#define RS_MASK (uint32_t)(0x1F<<RS_SHIFT)
#define RT_MASK (uint32_t)(0x1F<<RT_SHIFT)
#define RD_MASK (uint32_t)(0x1F<<RD_SHIFT)
#define SH_MASK (uint32_t)(0x1F<<SH_SHIFT)
#define FC_MASK (uint32_t)(0x3F<<FC_SHIFT)
//for I-Format
#define IM_MASK 0x0000FFFF
//for J-Format
#define TA_MASK 0x03FFFFFF
#define TA_SIGN_MASK (uint32_t)(pow(2,25))

//detection bit masks
#define R_FORMAT_DETECT	(uint32_t)(0x10<<OP_SHIFT)		//bin (010 000)
#define I_FORMAT_DETECT	(uint32_t)(0x2c<<OP_SHIFT)		//bin (101 100)
#define J_FORMAT_DETECT	(uint32_t)(0x02<<OP_SHIFT)		//bin (000 010)

//handy macros (stolen from solution to hw1 ;)
#define GETFIELD(val, mask, shift) (((val) & (mask)) >> (shift))

#define GETOP(val) GETFIELD(val, OP_MASK, OP_SHIFT)
#define GETRS(val) GETFIELD(val, RS_MASK, RS_SHIFT)
#define GETRT(val) GETFIELD(val, RT_MASK, RT_SHIFT)
#define GETRD(val) GETFIELD(val, RD_MASK, RD_SHIFT)
#define GETSH(val) GETFIELD(val, SH_MASK, SH_SHIFT)
#define GETFC(val) GETFIELD(val, FC_MASK, FC_SHIFT)
#define GETIM(val) ((val) & (IM_MASK))
#define GETTA(val) ((val) & (TA_MASK))

//some global vars
Instruction instructionMemory[1024], instructionRegister;
Address programCounter=0;

/* ---------------------------------------------------------------------------- */

/* Sign extens the target address of a J-Format Jump from 26 to 32 bits
*  params: in - the input
*  rets:   the extended 32 bit number
*/
int sign_Extend(uint32_t in){
	int i;

	for (i=1;i<7;i++)
		in |= ((in&TA_SIGN_MASK)<<i);
	return in;
}	
/* Reads an instruction from spec byte address
*  params:	addr - the byte address
*  rets:	the Instruction
*/
Instruction readInstructionMemory(Address addr){
	return instructionMemory[addr/4];
}


/* Writes an instruction to spec byte address
*  params:	addr	- the byte address
		ins 	- the Instruction
*  rets:	none
*/
Instruction writeInstructionMemory(Address addr, Instruction ins){
	instructionMemory[addr/4] = ins;
}

/* Input routine
*  Reads a line at a time from the stdin and puts it in Instruction Memory
*
*  params: 	none
*  rets:	none
*/
void read_input(){
	uint32_t address, command;

	while (EOF != scanf("%x %x",&address,&command))	
		writeInstructionMemory(address,command);
	
}

/* Decoding routine
*  Goes through the Instruction Mem and decodes (at least from our point of view) the instructions in there, one-by-one
*  outputs the actual assembly code for this instruction
*  params: 	none
*  rets:	none
*/
void decode()
{
	Formats form;
	int op,rs,rt,rd,shamt,funct,im;
	int32_t target;
	char output[100];
	int16_t t_im;
	
	//main loop - get the instruction
	while (0 != (instructionRegister = readInstructionMemory(programCounter))){
		//zeroize output string
		output[0] = 0; 
		
		//figure out the Format
		if ( !(instructionRegister&OP_MASK) || (instructionRegister & R_FORMAT_DETECT) ) form = R_FORM;
		else if (instructionRegister & I_FORMAT_DETECT) form = I_FORM;
		else if (instructionRegister & J_FORMAT_DETECT) form = J_FORM;
		else  continue;
		
		op = GETOP(instructionRegister);
		//go ahead and decode instruction:
		switch(form){
			case R_FORM:
				//get the different fields
				rs 		= GETRS(instructionRegister);
				rt 		= GETRT(instructionRegister);
				rd 		= GETRD(instructionRegister);
				shamt	= GETSH(instructionRegister);
				funct		= GETFC(instructionRegister);
				switch(funct){
					case ADD:
						sprintf(output,"add $%d,$%d,$%d",rd,rs,rt);
						break;
					case SUB:
						sprintf(output,"sub $%d,$%d,$%d",rd,rs,rt);
						break;
				}

				break;
				
			case I_FORM:
				rs 		= GETRS(instructionRegister);
				rt 		= GETRT(instructionRegister);
				im		= GETIM(instructionRegister);
				switch(op){
					case BEQ:
						//I dont need to "successfully sign-extent", coz this works too ;)
						t_im = (im & 0xFFFF);
						sprintf(output,"beq $%d,$%d,%d",rs,rt,(t_im*4));
						break;
					case ADDIU:
						sprintf(output,"addiu $%d,$%d,%d",rt,rs,im);
						break;
					case LW:
						sprintf(output,"lw $%d,%d($%d)",rt,im,rs);
						break;
				}

				break;
				
			case J_FORM:
				target	= sign_Extend(GETTA(instructionRegister));
				switch(op){
					case J:
						sprintf(output,"j %d",(target*4));
						break;
				}
				
				break;
		}
		//output all
		if (strlen(output) > 0)
			printf("%08x %08x %s\n",programCounter,instructionRegister,output);
		//inc PC
		programCounter += 4;
	
	}
}

/* main function
*  is called first and then just calls the two functions defined above
*  params: 	none
*  rets:	status-code (here:0)
*/
int main(){
	//read the input into instruction mem
	read_input();
	
	//and process
	decode();
		
	return 0;
}
